#ifdef USE_QT_WIDGETS
 #include <QtWidgets>
#endif

#include <QtGui>
#include <math.h>

#include "mandelbrotwidget.h"


const double DefaultCenterX = -0.637011f;
const double DefaultCenterY = -0.0395159f;
const double DefaultScale   = +0.00403897f;

const double ZoomInFactor   = 0.8f;
const double ZoomOutFactor  = 1.0 / ZoomInFactor;
const int ScrollStep = 20;

MandelbrotWidget::MandelbrotWidget (QWidget * parent) : QWidget (parent) {
    
  QIcon ico (":app.png");
  setWindowIcon (ico);
  setWindowTitle("Mandelbrot Image");
  
  centerX = DefaultCenterX;
  centerY = DefaultCenterY;
  pixmapScale = DefaultScale;
  curScale    = DefaultScale;

  qRegisterMetaType<QImage> ("QImage");
  connect (&thread, SIGNAL (renderedImage (QImage,double)), this, SLOT (updatePixmap (QImage,double)));

#ifndef QT_NO_CURSOR
  setCursor (Qt::CrossCursor);
#endif
  resize (1024, 1024);

}
void MandelbrotWidget::saveImage (void) {
  
  QRect rec = QApplication::desktop()->screenGeometry();
  qDebug ("Resolution = %d x %d", rec.width(), rec.height());

  QString fileName = QFileDialog::getSaveFileName (this,
               tr ("Save File"), "", "*.png");

  if (!fileName.isEmpty()) {
    QSize  isz (rec.width(), rec.height());
    QImage img (isz, QImage::Format_RGB32);
    thread.renderImage (img);
    QImageWriter writer(fileName, "png");
    writer.setText ("Author", "MandelbrotWidget");
    writer.write   (img);
    qDebug("File %s saved.", fileName.toLatin1().constData());
  }
  
}

void MandelbrotWidget::paintEvent (QPaintEvent * /* event */) {
  QPainter painter (this);
  painter.fillRect (rect(), Qt::black);

  if (pixmap.isNull()) {
    painter.setPen (Qt::white);
    painter.drawText (rect(), Qt::AlignCenter, tr ("Rendering initial image, please wait..."));
    return;
  }
  if (curScale == pixmapScale) {
    painter.drawPixmap (pixmapOffset, pixmap);
  } else {
    double scaleFactor = pixmapScale / curScale;
    int newWidth  = int (pixmap.width()  * scaleFactor);
    int newHeight = int (pixmap.height() * scaleFactor);
    int newX = pixmapOffset.x() + (pixmap.width()  - newWidth)  / 2;
    int newY = pixmapOffset.y() + (pixmap.height() - newHeight) / 2;

    painter.save();
    painter.translate (newX, newY);
    painter.scale (scaleFactor, scaleFactor);
    QRectF exposed = painter.matrix().inverted().mapRect (rect()).adjusted (-1, -1, 1, 1);
    painter.drawPixmap (exposed, pixmap, exposed);
    painter.restore();
  }
}

void MandelbrotWidget::resizeEvent (QResizeEvent * /* event */) {
  thread.render (centerX, centerY, curScale, size());
}

void MandelbrotWidget::keyPressEvent (QKeyEvent *event) {
  switch (event->key()) {
  case Qt::Key_Escape:  close(); break;
  case Qt::Key_F1: saveImage (); break; 
  case Qt::Key_Home:
    centerX = DefaultCenterX;
    centerY = DefaultCenterY;
    pixmapScale = DefaultScale;
    curScale = DefaultScale;
    thread.render (centerX, centerY, curScale, size());
    break;
  case Qt::Key_Plus:     zoom (ZoomInFactor);        break;
  case Qt::Key_Minus:    zoom (ZoomOutFactor);       break;
  case Qt::Key_Left:     scroll (-ScrollStep, 0);    break;
  case Qt::Key_Right:    scroll (+ScrollStep, 0);    break;
  case Qt::Key_Down:     scroll (0, -ScrollStep);    break;
  case Qt::Key_Up:       scroll (0, +ScrollStep);    break;
  case Qt::Key_0: thread.setPalette (0); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_1: thread.setPalette (1); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_2: thread.setPalette (2); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_3: thread.setPalette (3); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_4: thread.setPalette (4); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_5: thread.setPalette (5); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_6: thread.setPalette (6); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_7: thread.setPalette (7); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_8: thread.setPalette (8); thread.render (centerX, centerY, curScale, size()); break;
  case Qt::Key_9: thread.setPalette (9); thread.render (centerX, centerY, curScale, size()); break;
  default:
    QWidget::keyPressEvent (event);
  }
}

void MandelbrotWidget::wheelEvent (QWheelEvent *event) {
  int numDegrees = event->delta() / 8;
  double numSteps = numDegrees / 15.0f;
//zoom (pow (ZoomInFactor, numSteps));  // původně
  wheel (pow (ZoomInFactor, numSteps), event->pos());
}

void MandelbrotWidget::mousePressEvent (QMouseEvent *event) {
  if (event->button() == Qt::LeftButton)
    lastDragPos = event->pos();
}

void MandelbrotWidget::mouseMoveEvent (QMouseEvent *event) {
  if (event->buttons() & Qt::LeftButton) {
    pixmapOffset += event->pos() - lastDragPos;
    lastDragPos = event->pos();
    update();
  }
}

void MandelbrotWidget::mouseReleaseEvent (QMouseEvent *event) {
  if (event->button() == Qt::LeftButton) {
    pixmapOffset += event->pos() - lastDragPos;
    lastDragPos = QPoint();

    int deltaX = (width()  - pixmap.width())  / 2 - pixmapOffset.x();
    int deltaY = (height() - pixmap.height()) / 2 - pixmapOffset.y();
    scroll (deltaX, deltaY);
  }
}

void MandelbrotWidget::updatePixmap (const QImage &image, double scaleFactor) {
  if (!lastDragPos.isNull())
    return;

  pixmap = QPixmap::fromImage (image);
  pixmapOffset = QPoint();
  lastDragPos = QPoint();
  pixmapScale = scaleFactor;
  update();
}
void MandelbrotWidget::wheel(double zoomFactor, const QPoint & pos) {
  const double oldScale = curScale;
  curScale *= zoomFactor;
  update();
  const int hx = width() >> 1, hy = height() >> 1;
  const int dx = pos.x() - hx, dy = pos.y() - hy;
  const double ds = oldScale - curScale;
  centerX += dx * ds;
  centerY += dy * ds;
  update();
  thread.render (centerX, centerY, curScale, size());
}

void MandelbrotWidget::zoom (double zoomFactor) {
  curScale *= zoomFactor;
  update();
  thread.render (centerX, centerY, curScale, size());
}

void MandelbrotWidget::scroll (int deltaX, int deltaY) {
  centerX += deltaX * curScale;
  centerY += deltaY * curScale;
  update();
  thread.render (centerX, centerY, curScale, size());
}
