#include <QtGui>
#ifdef USE_QT_WIDGETS
#include <QtConcurrent>
#endif
#include <math.h>
#include "render.h"

static void processLine (QImage & image, const uint * colormap,
                         const double scaleFactor, const double centerX, const double centerY,
                         const int halfHeight, const int halfWidth, const int y) {

  uint * scanLine = reinterpret_cast<uint *> (image.scanLine (y));
  double ay = centerY + ( (y - halfHeight) * scaleFactor);

  const int MaxIterations = 128;
  const int Limit = 4;

  for (int x = -halfWidth; x < halfWidth; ++x) {
    double ax = centerX + (x * scaleFactor);
    int numIterations = 0;

    double zx=0,zy=0;
    do {
      ++numIterations;
      double a2 = zx * zx;
      double b2 = zy * zy;
      if ( (a2 + b2) > Limit) break;
      zy = 2.0 * zx * zy + ay;
      zx = a2 - b2 + ax;
    } while (numIterations < MaxIterations);

    if (numIterations < MaxIterations) {
      *scanLine++ = colormap[numIterations % ColormapSize];
    } else {
      *scanLine++ = qRgb (0, 0, 0);
    }
  }
}

void Wrap::operator() (const Chunk& chunk) {
  //printf ("%p: %04d-%04d\n", (void*) QThread::currentThreadId(), chunk.from, chunk.to);
  for (int y=chunk.from; y<chunk.to; y++)
    processLine (image, colormap, scaleFactor, centerX, centerY, halfHeight, halfWidth, y);
}

Render::Render (QObject * parent) : QObject (parent) {
  centerX = -0.637011f; centerY = -0.0395159f;
  scaleFactor = 0.0015;
  setPalette (2);
  running = false;

  connect (&watcher, SIGNAL (finished()), this, SLOT (handleFinished()));
}

Render::~Render() {

}
void Render::handleFinished() {
  emit renderedImage (mandelImage, scaleFactor);
  running = false;
}

void Render::setPalette (int palette) {
  union {
    struct {
      unsigned char r;
      unsigned char g;
      unsigned char b;
      unsigned char n;
    };
    unsigned color;
  } rgbn;
  for (int i = 0; i < ColormapSize; i++) {
    switch (palette) {
    case 0:
      rgbn.r=i<<1;
      rgbn.g=i<<1;
      rgbn.b=i<<1;
      break;
    case 1:
      rgbn.r=i*10.0;
      rgbn.g=127+127.0*sin (i/30.0);
      rgbn.b=127+127.0*sin (i/10.0);
      break;
    case 2:
      rgbn.r=127+127.0*sin (i/30.0);
      rgbn.g=i*10;
      rgbn.b=127+127.0*sin (i/10.0);
      break;
    case 3:
      rgbn.r=127+127.0*sin (i/30.0);
      rgbn.g=127+127.0*sin (i/10.0);
      rgbn.b=i*10.0;
      break;
    case 4:
      rgbn.r=127+127.0*sin (i/30.0);
      rgbn.g=i*20;
      rgbn.b=127+127.0*sin (i/10.0);
      break;
    case 5:
      rgbn.r=i*5;
      rgbn.g=255-i*7;
      rgbn.b=i*9;
      break;
    case 6:
      rgbn.r=127+127.0*sin (i/30.0);
      rgbn.g=i*20;
      rgbn.b=127-127.0*sin (i/30.0);
      break;
    case 7:
      rgbn.r=127-127.0*sin (i/30.0);
      rgbn.g=i*10;
      rgbn.b=127+127.0*sin (i/10.0);
      break;
    case 8:
      rgbn.r=127+127.0*sin (i/30.0);
      rgbn.g=127-127.0*sin (i/10.0);
      rgbn.b=i*10;
      break;
    case 9:
      rgbn.r=127-127.0*sin (i/10.0);
      rgbn.g=i*20;
      rgbn.b=127+127.0*sin (i/20.0);
      break;
    }
    rgbn.n = 0;

    colormap [i] = rgbn.color;
  }
}
void Render::render (double centerX, double centerY, double scaleFactor, QSize resultSize) {
  if (watcher.isRunning()) watcher.cancel();
  this->centerX = centerX;
  this->centerY = centerY;
  this->scaleFactor = scaleFactor;
  this->resultSize = resultSize;
  
  if (running) return;
  running = true;

  mandelImage = QImage (resultSize, QImage::Format_RGB32);
  renderImage (mandelImage);
}

bool Render::renderImage (QImage & image) {
  resultSize  = image.size();
  const int halfWidth  = resultSize.width()  / 2;
  const int Height     = resultSize.height();
  const int halfHeight = Height / 2;

  Wrap wrap (image, colormap, scaleFactor, centerX, centerY, halfHeight, halfWidth);

  chunks.clear();
  const int block = 64;
  int b = 0, e = 0;
  for (;;) {              // fill chunks
    e = b + block;
    if (e >= Height) e = Height - 1;
    if (b >= e) break;
    //qDebug("%d %d", b, e);
    Chunk ch (b,e);
    chunks.append (ch);
    b = e;
  }
  future = QtConcurrent::map (chunks, wrap);
  watcher.setFuture (future);
  future.waitForFinished();
  return false;
}

